{$A+,B-,D-,F-,G+,I+,K+,L-,N+,P+,Q-,R-,S+,T-,V-,W-,X+,Y-}
(******************************************************************************
**
**  TFILTERCALLBACK Component.
**
**  Author:   Andy Strong     e-mail: ANDREWS@NBAQSL.CO.UK
**                        CompuServe: 100716,3015
**
**  History
**  =======
**  Vers    Date      Comments
**  2.00  10 Aug 97   Added Blob Filtering.
**                    Added limited Property Editor - About.
**                    Changed TFilterRecord name to TFilterObject - more sensible.
**                    Discovered Paradox Tables can hold TDate(LongInt) as separate
**                      to TDateTime(Double).  Add fix to handle either correctly.
**  1.05  05 Aug 97   Added DbiFunction DbiGetExactRecordCount and
**                    FilteredRecordCount property.  NOTE: DO NOT REFERENCE THIS
**                    property from within the CALLBACK METHOD - YOU WILL CRASH
**                    YOUR SYSTEM!
**  1.04  24 Jul 97   Penultimate hack to fix AsDate and AsDateTime functions.
**                    Added AsTime method.
**  1.03  10 Jun 97   Ensure FieldList is loaded every time Active set to True.
**                    Ensuring list always upto date.
**  1.02  06 Feb 97   Fixes for Date/Time handling of raw data, and boolean
**                    handling.
**  1.01  Not Rel.    Minor Raw Data problems
**  1.00  10 Sep 96   First Public Release
**
**  A component to be dropped on a form, to allow low-level BDE callback
**  on a table independently of any other indexes, ranges set.
**  Encapsulates the BDE callback in a standard Delphi event passing
**  adequate information to allow selective filtering using Pascal code.
**
**  Taken from an idea on the BDE Compuserve forum.
**
**  This unit consists of three objects and one component.
**  The component is used to encapsulate the usage of the BDE callback filter,
**  and raw data handling.
**  The TFilterCallback Component sets up the callback filter for the BDE;
**  the TFilterObject object, is the object passed to the users Delphi callback
**  routine to enable easy access to the raw data to work on;
**  the TRaw object is an internal sanitiser of the raw BDE data for the Delphi
**  users event handler.
**
**  NOTE: The function DbiGetExactRecordcount works by iterating through all
**        records and requesting a filter callback to see if the records are
**        allowed in the set.  This means that the DB is potentially parsed
**        multiple times.
**        Once for the Filter in the first place.
**        Once for each time the DbiGetExcactRecordCount is called.
**        This may have a dramatic performance effect if BLOB filtering is being
**        used.
**
*******************************************************************************)
unit Fltcback;

interface

{$C PRELOAD FIXED}

{
  Remove this define if you want to compile without any dependancy upon the
  IDAPI01.DLL.  This will remove the ability to find out how many Records are
  returned by the Filter.  The FilteredRecordCount property calls an undocumanted
  BDE call.  This iterates through the filter mechanism again, just counting
  the number of 'Allows' = True.  Since this is the mechanism used, there can be
  a major performance impact, especially when filtering BLOB fields.
  The iteration is done EVERY time the DbiGetExactRecordCount is called, so if
  you are going to use it, minimise its use.
}
{$define BDEFilterCount}

uses
  Classes, DBTables, DB, DbiTypes, DbiProcs, SysUtils, DsgnIntf;

const
  MAXRAWSIZE = SizeOf(String);              { Max Pascal String size }
  VERSNUM = '2.00';

type
  ECallBackFilter = class(Exception);       { Special Filter Exception }
  TFilterObject = class;

  TBDEFieldType = (ftUnknown, ftString, ftDate, ftBlob, ftBoolean, ftSmallint,
    ftInteger, ftFloat, ftBCD, ftBytes, ftTime, ftDateTime, ftWord);

  {
    Object to handle the conversion of RAW data to a known Pascal type
  }
  TRawData = class(TObject)
  private
    FFldType: TBDEFieldType;                { Field Type }
  public
    RawData: array [0..MAXRAWSIZE] of byte; { Repository for BDE field raw data }
    function AsBoolean: boolean;
    function AsCurrency: Double;
    function AsDateTime: TDateTime;
    function AsDate : TDateTime;
    function AsFloat: Double;
    function AsInteger: LongInt;
    function AsSmallInt: Integer;
    function AsString: string;
    function AsTime: TDateTime;
  end;

  {
    Object which is used to hide all the nasty reading and seeking of the BDE
    blob data.  Encapsultes it into a read-only stream.
  }
  TFilterBlobStream = class(TStream)
  private
    FOpened: boolean;                       { Whether Blob field opened }
    FFieldNo: integer;                      { Field No in table }
    FPosition: Longint;                     { Position in Blob data }
    FFilterObject: TFilterObject;           { Reference to FilterObject }

    function GetBlobSize: Longint;
  public
    constructor Create(FieldNo: integer; FilterObject: TFilterObject);
    destructor Destroy; override;

    function Read(var Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: word): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;

    property Size: Longint read GetBlobSize;
    property Position: Longint read FPosition;
  end;

  {
    Object which is passed to the users Filter callback, giving access to the
    BDEs raw data for a record in a useable state.
  }
  TFilterObject = class(TObject)
  private
    FRawTable: TTable;                      { Pointer to Table being filtered }
    FFieldList: TStringList;                { List of All Field names in Table }
    FRaw: TRawData;                         { Pointer to RAW data handling object }
    FpRecBuf: pointer;                      { Pointer to BDE record structure }
    FPhysRecNum: Longint;                   { Physical record number }
    FReturnStream: TFilterBlobStream;       { Stream to return for blob data }

    procedure LoadFields(Table: TTable);    { Loads all Field Names into list }

  public

    property PhysicalRecordNum: Longint read FPhysRecNum;

    constructor Create;
    destructor Destroy; override;

    function GetBlobData(FieldName: openstring): TFilterBlobStream;
    function GetFieldData(FieldName: openstring): TRawData;
    function FieldIsBlank(FieldName: openstring): boolean;
    function IsValidBlobField(FieldName: openstring): boolean;
    function IsValidField(FieldName: openstring): boolean;
  end;

  {
    Function prototype for our sanitised Delphi filter event.
  }
  TFilterEvent = procedure(FilterRec: TFilterObject; var Allow: boolean) of object;

  {
    Actual component doing the work.
  }
  TFilterCallback = class(TComponent)
  private
    FActive: boolean;                       { Whether or not the filter is active }
    FFilterObject: TFilterObject;           { Pointer to our record handler }
    FTable: TTable;                         { Table Filter to be applied to }
    hFilter: hDbiFilter;                    { Handle of Filter when applied }
    FOnFilter: TFilterEvent;                { Users callback event handler }

    {$ifdef BDEFilterCount}
    function GetFilteredCount: Longint;
    {$endif}
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetActive(Value: boolean);
    procedure SetTable(Value: TTable);

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure StartFilter;
    procedure EndFilter;

  published
    property Active: boolean read FActive default False;
    {$ifdef BDEFilterCount}
    property FilteredRecordCount: Longint read GetFilteredCount;
    {$endif}
    property Table: TTable read FTable write SetTable;
    property OnFilter: TFilterEvent read FOnFilter write FOnFilter;
  end;

  {
    Little component editor to show an about box, and an About menu item on
    a right-click during design time.
  }
  TFilterEditor = class(TComponentEditor)
  public
    function GetVerb(Index: integer): string; override;
    function GetVerbCount: integer; override;
    procedure Edit; override;
    procedure ExecuteVerb(Index: Integer); override;
  end;

  procedure Register;

implementation

uses
  WinProcs, WinTypes, DbiErrs;

{$ifdef BDEFilterCount}
var
  LibHandle: THandle;
  DbiGetExactRecordCount: function (hCursor: hDBICur; var iRecCount: Longint): DBIResult;
{$endif}

(******************************************************************************
**  TRAWDATA Object functions/Methods/Procedures
******************************************************************************)
(*
**  Returns Raw data formatted as a boolean type
*)
function TRawData.AsBoolean: Boolean;
var
  PBool: ^WordBool;
begin
  PBool := @RawData;
  Result := PBool^;
end;

(*
**  Returns Raw data formatted as a Currency type
*)
function TRawData.AsCurrency: Double;
begin
  Result := AsFloat;
end;

(*
**  Returns Raw data formatted as DateTime type
*)
function TRawData.AsDateTime: TDateTime;
begin
  case FFldType of
    ftDate:
      Result := AsInteger;
    ftDateTime:
      Result := AsFloat  / MSecsPerDay;
    else
      Result := 0;
  end
end;

(*
**  Returns Raw data formatted as Date type
*)
function TRawData.AsDate: TDateTime;
begin
  case FFldType of
    ftDate:
      Result := AsInteger;
    ftDateTime:
      Result := Trunc(AsFloat  / MSecsPerDay);
    else
      Result := 0;
  end
end;

(*
**  Returns Raw data formatted as Time type
*)
function TRawData.AsTime: TDateTime;
begin
  case FFldType of
    ftTime:
      Result := AsInteger / MSecsPerDay;
    ftDateTime:
      Result := Frac(AsFloat / MSecsPerDay);
    else
      Result := 0;
  end
end;

(*
**  Returns Raw data formatted as a float type
*)
function TRawData.AsFloat: Double;
var
  PFloat: ^Double;
begin
  PFloat := @RawData;
  Result := PFloat^;
end;

(*
**  Returns Raw data formatted as a LongInt type
*)
function TRawData.AsInteger: LongInt;
var
  PInteger: ^LongInt;
begin
  PInteger := @RawData;
  Result := PInteger^;
end;

(*
**  Returns Raw data formatted as a pascal string type
*)
function TRawData.AsString: string;
begin
  Result := StrPas(@RawData);
end;

(*
**  Returns Raw data formatted as an integer type
*)
function TRawData.AsSmallInt: Integer;
var
  PInteger: ^Integer;
begin
  PInteger := @RawData;
  Result := PInteger^;
end;

(******************************************************************************
**  TFilterObject Object functions/Methods/Procedures
******************************************************************************)

(*
**  Constructor for client callback object
*)
constructor TFilterObject.Create;
begin
  inherited Create;                         { Create base object }
  FFieldList := TStringList.Create;         { Create Field Names Stringlist }
  FRaw := TRawData.Create;                  { Create Raw data handling object }
  FpRecBuf := nil;                          { initialise pointers to nil }
  FRawTable := nil;
  FReturnStream := nil;
end;

(*
**  Destructor for client callback object.  Release all associated objects
**  resources.
*)
destructor TFilterObject.Destroy;
begin
  FpRecBuf := nil;
  FRawTable := nil;
  FFieldList.Free;
  FFieldList := nil;
  FRaw.Free;
  FRaw := nil;
  if Assigned(FReturnStream) then
    FReturnStream.Free;                     { Free up stream if necessary }
  FReturnStream := nil;
  inherited Destroy;
end;

(*
**  Loads internal list with all field names in passed table.  The list is in
**  the same order as the fields are in the file definition, allowing the
**  list item number plus one to be used in the DbiGetField function.
*)
procedure TFilterObject.LoadFields(Table: TTable);
var
  CProps: CurProps;                         { Storage for table properties }
  FDescs: pFieldDescList;                   { Pointer to list of field descs }
  i: integer;
begin
  FFieldList.Clear;                         { Clear any old field descriptions }
  Check(DbiGetCursorProps(
    FRawTable.Handle, CProps));             { Get various info about table }
  GetMem(FDescs,
    CProps.iFields * SizeOf(FLDDesc));      { Create space for all field defs }
  try
    Check(DbiGetFieldDescs(FRawTable.Handle,
      pfldDesc(FDescs)));                   { Get all Field definitions }
    for i := 0 to CProps.iFields -1 do      { Iterate down all fields returned }
      FFieldList.AddObject(StrPas(FDescs^[i].szName),
        Pointer(FDescs^[i].iFldType));      { Add to list in field defined order }
  finally
    FreeMem(FDescs,
      CProps.iFields * SizeOf(FLDDesc));    { Free up field structure }
  end;
end;

(*
**  Function to decide whether passed field name exists in the table
*)
function TFilterObject.IsValidField(FieldName: openstring): boolean;
var
  Locn: integer;
begin
  Locn := FFieldList.IndexOf(FieldName);    { Find location in field list }
  Result := Locn >= 0;                      { in list if index >= zero }
end;

(*
**  Function to decide whether passed field name exists in the table
*)
function TFilterObject.IsValidBlobField(FieldName: openstring): boolean;
var
  Locn: integer;
begin
  Locn := FFieldList.IndexOf(FieldName);    { Find location in field list }
  Result := not ((Locn < 0) or
    (TBDEFieldType(FFieldList.Objects[Locn]) <> ftBlob)){ Returns -1 if not in list }
end;

(*
**  Function to decide whether the passed fields data is blank
*)
function TFilterObject.FieldIsBlank(FieldName: openstring): boolean;
var
  Blank: wordbool;
begin
  Result := True;                           { Assume it is blank }
  if IsValidField(FieldName) and
    (FRawTable <> nil) then                 { Check it exists! }
  begin
    Check(DbiGetField(FRawTable.Handle,
      FFieldList.IndexOf(FieldName) + 1,
      FpRecBuf, nil, Blank));               { Get status with exception checking }
    Result := Blank;                        { Return result }
  end;
end;

(*
**  Function to get Raw data for passed field name.
*)
function TFilterObject.GetFieldData(FieldName: openstring): TRawData;
var
  Blank: wordbool;
  FieldNo: integer;
begin
  Result := FRaw;                           { Point at our Raw data handler }
  if (FRawTable = nil) then
    FillChar(FRaw, MAXRAWSIZE, 0)           { Clear out raw data }
  else if IsValidField(FieldName) then      { If valid field name }
  begin
    FieldNo := FFieldList.IndexOf(FieldName);
    Check(DbiGetField(FRawTable.Handle, FieldNo + 1,
      FpRecBuf, @(Result.RawData), Blank)); { Get data with exception handling }
    Result.FFldType := TBDEFieldType(LongInt(FFieldList.Objects[FieldNo]));
  end else
    raise ECallBackFilter.Create(FieldName + ' not in Table')
end;

(*
**  Function to get Raw data for passed field name.
*)
function TFilterObject.GetBlobData(FieldName: openstring): TFilterBlobStream;
var
  Blank: wordbool;
begin
  Result := nil;
  if Assigned(FRawTable) then
  begin
    if Assigned(FReturnStream) then         { Already used Blob in this filter? }
      FReturnStream.Free;                   { yes - release it }
    FReturnStream := nil;
    if IsValidBlobField(FieldName) then     { If valid field name }
    begin
      FReturnStream := TFilterBlobStream.   { Create Blob Stream }
        Create(FFieldList.IndexOf(FieldName) + 1, Self);
      Result := FReturnStream;              { Return stream result }
    end else
      raise ECallBackFilter.Create(FieldName + ' not in Table, or NOT blob field')
  end
end;

(******************************************************************************
**  TFILTERBLOBSTREAM Component functions/Methods/Procedures
******************************************************************************)
(*
**  Filter routine for reading blob fields.  These are presented as streams.
**  When the GetBlobData is used, a temporary stream is created to the Fields
**  Blob data (an exception occurs if the field is not a blob field).  The stream
**  is released before the next record is passed in by the filter mechanism.  The
**  previous stream is also released if the GetBlobData call is issued again in
**  same filter routine.  This makes it possible to check multiple blob fields
**  data in the same routine.  However, only ONE may be open at any one time.
*)

(*
**  Attempts to create a stream into the blob data of the field specified.
**  If successful, sets FOpened flag.
*)
constructor TFilterBlobStream.Create(FieldNo: integer; FilterObject: TFilterObject);
begin
  FOpened := False;                         { Blob Field not opened yet }
  FPosition := 0;                           { Default to start of data }
  FFieldNo := FieldNo;                      { Save field no. }
  FFilterObject := FilterObject;            { Pointer to Field struct data }
  with FFilterObject do
    Check(DbiOpenBlob(FRawTable.Handle,
      FpRecBuf, FFieldNo, dbiReadOnly));    { Exception if open fails }
  FOpened := True;                          { Gets here if opened OK }
end;

(*
**  Destructor for the FilterBlobStream object.  If the Blob data was opened
**  (it may fail during the constructor), free the blob field resources.
*)
destructor TFilterBlobStream.Destroy;
begin
  if FOpened then
    with FFilterObject do
    Check(DbiFreeBlob(FRawTable.Handle, FpRecBuf, FFieldNo));
  inherited Destroy;
end;

procedure __AHSHIFT; far; external 'KERNEL' index 113;
(*
**  Function to skip a pointer by OFS bytes, covering segment boundary
**  skipping.
*)
function OffsetPointer(P: Pointer; Ofs: Longint): Pointer; assembler;
asm
        MOV     AX,Ofs.Word[0]
        MOV     DX,Ofs.Word[2]
        ADD     AX,P.Word[0]
        ADC     DX,0
        MOV     CX,OFFSET __AHSHIFT
        SHL     DX,CL
        ADD     DX,P.Word[2]
end;

(*
**  Reads the specified amount of data from the blob field into the passed
**  buffer area.  We need to be careful that we don't cross a segment boundary.
*)
function TFilterBlobStream.Read(var Buffer; Count: Longint): Longint;
const
  CHUNKSIZE = 32767;
var
  PBuff: Pointer;
  Chunk: word;
  BytesRead: Longint;
  Status: DbiResult;
begin
  PBuff := @Buffer;                         { Point at start of passed buffer }
  Result := 0;
  while Count > 0 do
  begin
    Chunk := CHUNKSIZE;                     { Read in default chunk size }
    if Chunk > Count then
      Chunk := Count;                       { Only read in waht we need to }
    with FFilterObject do
    Status := DbiGetBlob(FRawTable.Handle, FpRecBuf, FFieldNo, FPosition,
      Chunk, PBuff, BytesRead);
    case Status of
      DBIERR_NONE,
      DBIERR_ENDOFBLOB:
        Inc(Result, BytesRead);
      DBIERR_INVALIDBLOBOFFSET:             { Valid Statuses }
        ;
    else                                    { Other error - report it }
      DbiError(Status);
    end;
    if Status <> DBIERR_NONE then
      Break;                                { End of WHILE loop, but not error }
    Inc(FPosition, BytesRead);              { Bump Blob Pointer }
    Dec(Count, BytesRead);                  { Decrement number of bytes read }
    PBuff := OffsetPointer(PBuff, BytesRead);{ Next block in User Buffer }
  end
end;

(*
**  Changes the current position pointer within the blob data by Offset amount.
**  If Origin is:
**    0 - Offset is relative to the beginning of the blob data
**    1 - Offset is relative to the current position in the blob data
**    2 - Offset is relative to the end of the blob data
*)
function TFilterBlobStream.Seek(Offset: Longint; Origin: word): Longint;
begin
  if FOpened then
    case Origin of
      0:
        FPosition := Offset;                { Absolute position }
      1:
        Inc(FPosition, Offset);             { Relative to where we are now }
      2:
        FPosition := Size + Offset;         { Relative to the end of the data }
    end;
  Result := FPosition;
end;

(*
**  Stops attempts at writing to our filtered stream!
*)
function TFilterBlobStream.Write(const Buffer; Count: Longint): Longint;
begin
  raise ECallbackFilter.Create('Filtered Blobs are ReadOnly');
end;

(*
**  Returns the size of the current blob field
*)
function TFilterBlobStream.GetBlobSize: Longint;
begin
  Result := -1;
  if FOpened then
  with FFilterObject do
    Check(DbiGetBlobSize(FRawTable.Handle, FpRecBuf, FFieldNo, Result));
end;

(******************************************************************************
**  TFILTERCALLBACK Component functions/Methods/Procedures
******************************************************************************)

(*
**  Function which is called by the low level BDE routine for each record to be
**  filtered.  This function must return True to allow Delphi to see the record.
**  When we setup the callback in SETACTIVE, we passed a pointer to our
**  instantiation of TFILTERCALLBACK.  This is used in this isolated function
**  to access the users callback function, and our data handling object.
*)
function BDEFilterCallback(ulClientdata: Longint; pRecBuff: Pointer;
          iPhyRecNum: Longint): Integer; export;
var
  Allow: boolean;
begin
  Allow := True;                            { Allow through this filter? }
  with TFilterCallBack(ulClientdata) do     { Pointer to our TFilterCallback Obj }
  if Assigned(FOnFilter) then               { User callback function assigned? }
  begin                                     { Yes... }
    FFilterObject.FpRecBuf := pRecBuff;     { Pass current record to handler obj }
    FFilterObject.FPhysRecNum := iPhyRecNum;{ Pass physical record No. to handler obj }
    FOnFilter(FFilterObject, Allow);        { Call user filter function }
    if Assigned(FFilterObject.FReturnStream) then { Blob stream opened? }
    begin
      FFilterObject.FReturnStream.Free;     { Must release it before we move on }
      FFilterObject.FReturnStream := nil;   { Flag as not used }
    end
  end;
  Result := Integer(Allow);                 { Function has decided to allow record }
end;

(*
**  Create component which handles the BDE callbacks.
*)
constructor TFilterCallBack.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FOnFilter := nil;
  FTable := nil;
  hFilter := nil;
  FActive := False;
  FFilterObject := TFilterObject.Create;    { Create client record handler }
end;

(*
**  Destroy resources used by component
*)
destructor TFilterCallBack.Destroy;
begin
  if FActive then
    SetActive(False);                       { Disable filter }
  FOnFilter := nil;                         { Stop GPFs }
  FFilterObject.Free;
  inherited Destroy;
end;

(*
**  Starts filter active
*)
procedure TFilterCallBack.StartFilter;
begin
  SetActive(True);
end;

(*
**  Terminates filter
*)
procedure TFilterCallBack.EndFilter;
begin
  SetActive(False);
end;

(*
**  Activate or Deactivate the filter callback.
*)
procedure TFilterCallBack.SetActive(Value: boolean);
begin
  if (Value <> FActive) then
  begin
    FActive := Value;                         { Set value }
    if not (csDesigning in ComponentState)
      and (FTable <> nil) then
    with FTable do
    begin
      try
        DisableControls;                      { Un-link any data controls }
        if FFilterObject.FFieldList.Count < 1 then{ Fields not yet loaded... }
          FFilterObject.LoadFields(FTable);   { Load Table Fields }
        if Value then                         { Activating }
        begin
          Check(DbiAddFilter(Handle, Longint(Self),
            1, False, nil, BDEFilterCallBack,
            hFilter));                        { Set filter to pass pointer to us }
          Check(DbiActivateFilter(
            Handle, hFilter));                { Initiate filter }
        end else
        begin
          Check(DbiDeActivateFilter(
            Handle, hFilter));                { Deactivate our filter }
          Check(DbiDropFilter(Handle, hFilter));{ Remove it from table }
          FFilterObject.FFieldList.Clear;     { Clear old Field definitions }
        end;
        First;                                { Reset table }
      finally
        EnableControls;                       { Re-Link data controls }
      end
    end
  end;
end;

(*
**  When a table name is entered, read in all field names.
**  Whilst then component is created, the LoadFields method cannot be called
**  because the dependant variables and objects have not been completely
**  created.  This eventuality is taken care of by the SetActive method
*)
procedure TFilterCallBack.SetTable(Value: TTable);
begin
  if (Value <> FTable) then
  begin
    FTable := Value;                        { Save New Table }
    FFilterObject.FRawTable := Value;       { Save user referenced pointer }
    FFilterObject.FFieldList.Clear;         { Clear old Field definitions }
    if not ((csDesigning in ComponentState) or
      (csLoading in ComponentState)) then
      FFilterObject.LoadFields(Value);      { Only Fill fields at run-time }
  end;
end;

(*
**  Removes our reference to the selected table if it is being removed
*)
procedure TFilterCallBack.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if (AComponent = FTable) and (Operation = opRemove) then
  begin
    SetActive(False);                       { Disable filter first }
    FFilterObject.FFieldList.Clear;         { Clear old Field definitions }
    FTable := nil;
    FFilterObject.FRawTable := nil;         { Save user referenced pointer }
  end
end;

{$ifdef BDEFilterCount}
(*
**  Returns the exact amount of records after the filter has been applied
*)
function TFilterCallBack.GetFilteredCount: Longint;
begin
  Result := -1;
  if Assigned(FTable) then                  { Table assigned? }
    if FActive then                         { Filter Active? }
      Check(DbiGetExactRecordCount(FTable.Handle, Result))
    else
      Result := FTable.RecordCount;         { No filter, return normal count }
end;
{$endif}

(******************************************************************************
**  TFILTEREDITOR Component editor functions/Methods/Procedures
******************************************************************************)

(*
**  Returns number of entries to be added to the components right click menu
*)
function TFilterEditor.GetVerbCount: integer;
begin
  Result := 1;
end;

(*
**  Returns the strings to be loaded into the right click menu for each of the
**  GETVERBCOUNT indexes.
*)
function TFilterEditor.GetVerb(Index: integer): string;
begin
  case Index of
    0:
      Result := 'About FilterCallback';
  end;
end;

(*
**  What to do if the component is double-clicked upon
*)
procedure TFilterEditor.Edit;
begin
  ExecuteVerb(0);
end;

(*
**  What to do if the components right click menu item is selected
*)
procedure TFilterEditor.ExecuteVerb(Index: integer);
begin
  if Index = 0 then                         { Only respond to About menu item }
    MessageBox(0, 'TFilterCallback - Version ' + VERSNUM + #13#10#13#10 +
                  'Delphi 1.0 Low level BDE callback component'#13#10 +
                  'Developed by Andy Strong: andys@nbaqsl.co.uk'#13#10#13#10 +
                  'This component should not be needed under Delphi 2'#13#10 +
                  'because of the TTable.OnFilter property.',
                  'About', MB_TASKMODAL or MB_OK or MB_ICONINFORMATION);
end;

(*
**  Standard Component Palette register procedure
*)
procedure Register;
begin
  RegisterComponents('Custom', [TFilterCallBack]);
  RegisterComponentEditor(TFilterCallback, TFilterEditor);
end;

{$ifdef BDEFilterCount}
(*
**  Unit Exit procedure to release one lock on the IDAPI01 DLL.
*)
procedure IDAPIExitProc;far;
begin
  if LibHandle >= 32 then                   { Was library loaded }
    FreeLibrary(LibHandle);
  LibHandle := 0;
end;

(*
**  Initialisation of the BWCC library DLL.  Loads the DLL into Memory and
**  gets the function pointers for the required calls.  If either the load
**  of the library, or the acquisition of any of the function pointers fails
**  we will raise an Exception.
**  NOTE: the second parameter of the GETPROCADDRESS function calls looks
**  strange:  the function normally expects a PCHAR pointing to the function
**  name that is required;  however we cannot retrieve the pointer by this
**  method, we must use the INDEX of the function.  Casting the number to a
**  PCHAR is the only way to get the compiler to pass the syntax.
**  Indexes to function in BWCC found using Borlands IMPDEF program.
*)
procedure Initialise;

  (*
  **  Local function to validate that an addres was retrieved.  If one wasn't
  **  we will not return!
  *)
  function ChkAddr(ProcAddr: TFarProc): TFarProc;
  begin
    Result := ProcAddr;                     { Setup return value }
    if not Assigned(ProcAddr) then          { Didn't find Procedure/function }
      raise Exception.Create('Function pointer not found in DLL!');
  end;

begin
  if LibHandle = 0 then                     { If already loaded - ignore }
  begin
    LibHandle :=
      WinProcs.LoadLibrary('IDAPI01');      { Attempt to load DLL }
    if LibHandle < 32 then
      raise Exception.Create('Could NOT load Library: IDAPI01.DLL');
    {
      Load all the function pointers from the DLL into our function prototypes
    }
    try
      @DbiGetExactRecordCount:= ChkAddr(GetProcAddress(LibHandle, PChar(883)));
    except
      FreeLibrary(LibHandle);               { Release Loaded Library }
      LibHandle := 0;                       { Flag as NO LONGER loaded }
      raise;                                { Re-raise }
    end;
  end;
end;

initialization
  LibHandle := 0;                           { Log as not yet loaded }
  Initialise;                               { Load Library and function pointers }
  AddExitProc(IDAPIExitProc);               { Ensure library unloaded if loaded }
{$endif}
end.
